# Copyright 2022, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import io
import json
import os
import sys
import threading

from json import JSONDecodeError

from digi.ccble.exceptions import BluetoothNotSupportedException, ConnectCoreBLEException
from digi.ccble.service import ConnectCoreBLEService

# Constants.
DIR_INTERFACES = "interfaces"

ERROR_REQUEST_FORMAT = "Invalid request format: '{}' not specified"

EXTENSION_TEXT = ".txt"

OPERATION_READ = "read"
OPERATION_WRITE = "write"

SETTING_SEPARATOR = "="

STATUS_ERROR = "error"
STATUS_OK = "ok"

TAG_DESCRIPTION = "desc"
TAG_INTERFACE = "interface"
TAG_OPERATION = "operation"
TAG_SETTING = "setting"
TAG_STATUS = "status"
TAG_VALUE = "value"

# Variables.
cc_ble_service = None
interfaces = {}


def data_received_callback(data):
    """
    Callback to be notified when new data is received from the connected device.

    After parsing the data, this method will perform a different action depending on the requested operation:
        - If it is a read request, it will send back the configuration data of the specified interface file.
        - If it is a write request, it will update the configuration data of the specified interface file.

    Args:
        data (Bytearray): received (JSON) data.
    """
    # Parse data as JSON.
    try:
        json_request = json.loads(data.decode("utf-8"))
    except JSONDecodeError:
        print("Error: could not parse JSON data")
        return
    # Check if the JSON data is correct.
    try:
        check_json(json_request)
    except ValueError as exc:
        # Error found during the JSON validation. Send error.
        send_error(str(exc))
        return
    # Process the request.
    process_request(json_request)


def check_json(json_properties):
    """
    Checks if json_properties follows the correct format.

    The defined format is as follows:
    {
        "operation"      : "read" / "write",        required: always
        "interface"      : "<interface_name>",      required: always
        "setting"        : "<setting_name>",        required: always
        "value"          : "<setting_value>",       required: Operation=write
    }

    Args:
        json_properties (Dictionary<String, String>): JSON string to be checked.

    Raises:
        ValueError: if the JSON string is not valid.
    """
    if TAG_OPERATION not in json_properties:
        raise ValueError(ERROR_REQUEST_FORMAT.format(TAG_OPERATION))
    if TAG_INTERFACE not in json_properties:
        raise ValueError(ERROR_REQUEST_FORMAT.format(TAG_INTERFACE))
    if TAG_SETTING not in json_properties:
        raise ValueError(ERROR_REQUEST_FORMAT.format(TAG_SETTING))
    if json_properties[TAG_OPERATION] == OPERATION_WRITE:
        if TAG_VALUE not in json_properties:
            raise ValueError(ERROR_REQUEST_FORMAT.format(TAG_VALUE))


def process_request(json_request):
    """
    Processes the request.

    Args:
        json_request (Dictionary<String, String>): JSON string to be processed.
    """
    # Retrieve the operation to execute.
    operation = json_request[TAG_OPERATION]
    # Retrieve the network interface.
    interface = json_request[TAG_INTERFACE]
    # Check if the interface exists.
    if interface not in interfaces:
        send_error(f"Interface '{interface}' not found")
        return
    # Retrieve the setting.
    setting = json_request[TAG_SETTING]
    # Check if the setting exists in the requested interface.
    if setting not in interfaces[interface]:
        send_error(f"Setting '{setting}' not found in interface '{interface}'")
        return
    # Execute the request.
    if operation == OPERATION_READ:
        json_response = json.dumps({TAG_STATUS: STATUS_OK,
                                    TAG_INTERFACE: interface,
                                    TAG_SETTING: setting,
                                    TAG_VALUE: interfaces[interface][setting]})
        send_data(json_response.encode('utf-8'))
    elif operation == OPERATION_WRITE:
        # Save previous value.
        old_value = interfaces[interface][setting]
        # Set new value.
        interfaces[interface][setting] = json_request[TAG_VALUE]
        # Save new value.
        try:
            save_interface(interface)
            json_response = json.dumps({TAG_STATUS: STATUS_OK})
            send_data(json_response.encode('utf-8'))
        except Exception as exc:
            send_error(f"Could not write setting: {str(exc)}")
            # Restore previous value.
            interfaces[interface][setting] = old_value
            return
    else:
        # If the request is not a read or write request, send an error response.
        send_error(f"Invalid operation '{operation}'")


def send_error(error_message):
    """
    Sends an error message to the client.

    Args:
        error_message (String): error message to be sent.
    """
    json_response = json.dumps({TAG_STATUS: STATUS_ERROR,
                                TAG_DESCRIPTION: error_message})
    send_data(json_response.encode('utf-8'))


def connection_callback():
    """
    Callback to be notified when a new connection is established.
    """
    print(f"Connection established using '{cc_ble_service.get_interface_type().title}'")


def send_data(data):
    """
    Sends data to the connected device through the Bluetooth Low Energy interface.

    Args:
        data (Bytearray): data to be sent.
    """
    try:
        cc_ble_service.send_data(data)
    except ConnectCoreBLEException as exc:
        print(f"Error: could not send data to connected device: {str(exc)}")


def save_interface(interface):
    """
    Saves the given interface.

    Args:
        interface (String): name of the interface to save.
    """
    with io.open(os.path.join(DIR_INTERFACES, interface + EXTENSION_TEXT),
                 'r+', encoding='utf-8') as file:
        # Overwrite file data.
        file.seek(0)
        file.truncate(0)
        for setting in interfaces[interface]:
            file.write(setting + SETTING_SEPARATOR + interfaces[interface][setting] + "\n")


def initialize_interfaces():
    """
    Initializes the interfaces
    """
    # Iterate all files of the interfaces directory.
    for file in os.listdir(DIR_INTERFACES):
        # Check if the file is a text file.
        if file.endswith(EXTENSION_TEXT):
            # Get the interface name.
            interface_name = file.split('.')[0]
            # Initialize interface settings.
            interface_settings = {}
            # Read the interface settings.
            with io.open(os.path.join(DIR_INTERFACES, file), 'r', encoding='utf-8') as iface_file:
                for line in iface_file:
                    line_dict = line.strip().split(SETTING_SEPARATOR)
                    interface_settings[line_dict[0]] = line_dict[1]
            # Add the interface to the list of interfaces.
            interfaces[interface_name] = interface_settings


def main():
    global cc_ble_service

    # Initialize the interfaces.
    try:
        initialize_interfaces()
    except Exception as exc:
        print(f"Error: could not initialize interfaces: {str(exc)}")
        sys.exit(1)
    # Initialize the service.
    try:
        cc_ble_service = ConnectCoreBLEService.get_instance()
    except BluetoothNotSupportedException:
        print("The system does not support bluetooth. Aborting...")
        sys.exit(1)
    # Register service callbacks.
    cc_ble_service.add_data_received_callback(data_received_callback)
    cc_ble_service.add_connect_callback(connection_callback)
    # Configure thread for service execution.
    application_thread = threading.Thread(target=cc_ble_service.start, daemon=True)
    application_thread.start()
    # Wait for input to stop application.
    print('Service started, hit <ENTER> to stop and exit.')
    try:
        input('> ')
        # Stop the service.
        print('Stopping the service...')
        try:
            cc_ble_service.stop()
            application_thread.join()
            application_thread = threading.Thread(target=cc_ble_service.start, daemon=True)
            print('Service stopped!')
        except ConnectCoreBLEException as exc:
            print(str(exc))
    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == '__main__':
    main()



# Copyright 2022, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import glob
import logging
import threading

from abc import ABC, abstractmethod
from enum import Enum

from bluezero.adapter import Adapter, AdapterError
from dbus.exceptions import DBusException

from serial.serialutil import SerialException
from serial.tools.list_ports_linux import SysFS
from serial.tools import list_ports

from digi.ccble import utils
from digi.ccble.exceptions import ConnectCoreBLEException, BluetoothNotSupportedException
from digi.ccble.peripheral import ConnectCoreBLEPeripheral
from digi.ccble.security import SRPSecurityManager
from digi.xbee.devices import XBeeDevice
from digi.xbee.exception import InvalidOperatingModeException, InvalidPacketException, XBeeException
from digi.xbee.models.atcomm import ATStringCommand
from digi.xbee.models.hw import HardwareVersion
from digi.xbee.models.mode import OperatingMode
from digi.xbee.models.options import XBeeLocalInterface
from digi.xbee.models.status import ModemStatus, ATCommandStatus
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.packets.base import UnknownXBeePacket, XBeeAPIPacket
from digi.xbee.packets.common import ATCommPacket, ATCommResponsePacket
from digi.xbee.packets.relay import UserDataRelayPacket, UserDataRelayOutputPacket

# Constants.
_ERROR_BLUETOOTH_NOT_SUPPORTED = "Bluetooth is not supported on this device."
_ERROR_CHANGE_ADVERTISE_NAME = "Failed to change advertise name: %s"
_ERROR_CHANGE_ENCRYPTION_PASSWORD = "Failed to change encryption password: %s"
_ERROR_READ_ADVERTISE_NAME = "Failed to read advertise name: %s"
_ERROR_SEND_DATA = "Failed to send data: %s"
_ERROR_START_BLUETOOTH_SERVICE = "Failed to start Bluetooth service: %s"
_ERROR_STOP_BLUETOOTH_SERVICE = "Failed to stop Bluetooth service: %s"

_WARNING_DATA_DECRYPT = "Could not decrypt received data %s."
_WARNING_DATA_ENCRYPT = "Could not encrypt data to send %s."
_WARNING_INVALID_PACKET_RECEIVED = "Invalid packet received: %s"
_WARNING_SPURIOUS_LOGIN_PACKET_RECEIVED = "Spurious login packet received"

_XBEE_API_FRAME_SRP_LOGIN_REQUEST = 0x2C
_XBEE_API_FRAME_SRP_LOGIN_ANSWER = 0xAC
_XBEE_PORT_PREFIXES = ["ttyXBee", "ttymxc", "ttymca", "ttyUSB"]
_XBEE_PORT_BAUDRATE = 115200


class BLEInterface(Enum):
    """
    Enumeration class listing all the Bluetooth Low Energy interfaces.
    """
    ANY = (0, "Any", "Use first available interface, starting with native adapters")
    ADAPTER = (1, "Bluetooth adapter", "Use native Bluetooth adapter interface")
    XBEE3 = (2, "XBee 3", "Use XBee 3 Bluetooth interface")

    def __init__(self, code, title, description):
        """
        Class constructor. Instantiates a new `BLEInterface` entry with the provided parameters.

        Args:
            code (Integer): BLE interface code.
            title (String): BLE interface title.
            description (String): BLE interface description.
        """
        self.__code = code
        self.__title = title
        self.__description = description

    @property
    def code(self):
        """
        Returns the BLE interface code.

        Returns:
            Integer: BLE interface code.
        """
        return self.__code

    @property
    def title(self):
        """
        Returns the BLE interface title.

        Returns:
            String: BLE interface title.
        """
        return self.__title

    @property
    def description(self):
        """
        Returns the BLE interface description.

        Returns:
            String: BLE interface description.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the BLE interface corresponding to the given code.

        Args:
            code (Integer): BLE interface code.

        Returns:
            SrpError: BLE interface corresponding to the given code, `None` if not found.
        """
        for ble_interface in cls:
            if code == ble_interface.code:
                return ble_interface
        return None


BLEInterface.__doc__ += utils.doc_enum(BLEInterface)


def get_bluetooth_adapter():
    """
    Returns the first available Bluetooth adapter in the system.

    Returns:
        :class:`.Adapter`: The first :class:`.Adapter` available in the system, `None` if
                           no adapter is found.
    """
    try:
        for bt_adapter in Adapter.available():
            if bt_adapter.powered:
                return bt_adapter
    except (AdapterError, DBusException):
        return None

    return None


def get_xbee_device():
    """
    Returns the first available XBee BTLE capable device in the system.

    Returns:
        :class:`.XBeeDevice`: The first BTLE :class:`.XBeeDevice` available in the system, `None`
                              if no device is found.
    """
    # List serial ports.
    serial_ports = list(list_ports.comports())
    # Add custom ports.
    devices = ['/dev/ttyXBee']
    devices.extend(glob.glob('/dev/ttymca*'))
    devices.extend(glob.glob('/dev/ttymxc*'))
    devices.extend(glob.glob('/dev/ttyLP*'))
    serial_ports.extend(info
                        for info in [SysFS(d) for d in devices]
                        if info.subsystem != "platform")
    # Iterate all ports found.
    for port in serial_ports:
        # Check whether port matches supported prefixes.
        if any(port.name.startswith(prefix) for prefix in _XBEE_PORT_PREFIXES):
            # Try to open the port with default baudrate and forcing settings.
            try:
                xbee_device = XBeeDevice(port.device, _XBEE_PORT_BAUDRATE)
                xbee_device.open(force_settings=True)
            except (InvalidOperatingModeException, XBeeException, SerialException):
                # Continue with next port.
                continue
            try:
                # Try to read Bluetooth setting, if exception is thrown then the device
                # is not BTLE capable.
                xbee_device.get_parameter(ATStringCommand.BT.command)
                # Try to read the SRP salt value, if exception is thrown then the device
                # is not SRP capable.
                xbee_device.get_parameter(ATStringCommand.DOLLAR_S.command)
                # Disable bluetooth to start from a known state.
                xbee_device.disable_bluetooth()
            except XBeeException:
                xbee_device.close()
                # Continue with next port.
                continue
            # The device is valid.
            return xbee_device
    # No devices found.
    return None


class ConnectCoreBLEService(ABC):
    """
    This class offers several methods that can be used to communicate a ConnectCore device
    working as a Peripheral device with an external device via Bluetooth Low Energy.

    The Bluetooth Low Energy communication can be based on native Bluetooth adapters or on the
    XBee 3 Bluetooth Low Energy interface.
    """

    # Variables.
    __BLE_service_instance = None  # Track a unique instance for the service (singleton).
    _log = logging.getLogger(__name__)

    def __init__(self, interface_type):
        """
        Class constructor. Instantiates a new :class:`.ConnectCoreBLEService` object with the given
        parameters.

        Args:
            interface_type (:class:`.BLEInterface`): the BLE interface type to use.
        """
        # Initialize variables.
        self._on_connect = []
        self._on_disconnect = []
        self._on_data_received = []
        self._interface_type = interface_type
        self.service_active = False

    @classmethod
    def get_instance(cls, ble_interface=BLEInterface.ANY):
        """
        Checks the available Bluetooth interfaces of the device and generates a single instance of
        the class depending on the preferred interface. If an instance already exist, the method
        simply returns that instance.

        Depending on the preferred interface, the method returns as follows:
          - If the interface is set to 'ANY' or 'ADAPTER' and Bluetooth is natively supported in
            the system, a :class:`.ConnectCoreBLEServiceNative` object is returned.
          - If the interface is set to 'ANY' and the system does not support native Bluetooth or
            the interface is set to 'XBEE3', the method tries to find a BLE capable XBee 3 device
            connected to the system and returns a :class:`.ConnectCoreBLEServiceXBee` object.

        Args:
            ble_interface (:class:`.BLEInterface`): the BLE interface type to use.

        Returns:
            :class:`.ConnectCoreBLEService`: the single service instance.

        Raises:
            BluetoothNotSupportedException: if the system does not have any valid Bluetooth
                                            interface.
        """
        # Check if an instance of same type already exists.
        if (cls.__BLE_service_instance is not None
                and ((cls.__BLE_service_instance.get_interface_type() == BLEInterface.ADAPTER
                      and ble_interface in (BLEInterface.ANY, BLEInterface.ADAPTER))
                     or (cls.__BLE_service_instance.get_interface_type() == BLEInterface.XBEE3
                         and ble_interface in (BLEInterface.ANY, BLEInterface.XBEE3)))):
            return cls.__BLE_service_instance
        # Initialize variables.
        adapter = None
        cls.__BLE_service_instance = None
        # Check the requested interface type.
        if ble_interface in (BLEInterface.ANY, BLEInterface.ADAPTER):
            # Check for native Bluetooth interface.
            adapter = get_bluetooth_adapter()
        if adapter is not None:
            cls.__BLE_service_instance = ConnectCoreBLEServiceNative(adapter.address)
            cls._log.info("Native Bluetooth interface found: %s", adapter.address)
        elif ble_interface in (BLEInterface.ANY, BLEInterface.XBEE3):
            # Check for XBee Bluetooth interface.
            xbee_device = get_xbee_device()
            if xbee_device is not None:
                cls.__BLE_service_instance = BLEServiceXBee(xbee_device)
                cls._log.info("XBee Bluetooth interface found: %s", xbee_device.get_64bit_addr())
        # If no available interface is found raise exception.
        if cls.__BLE_service_instance is None:
            cls._log.error(_ERROR_BLUETOOTH_NOT_SUPPORTED)
            raise BluetoothNotSupportedException()
        return cls.__BLE_service_instance

    def is_running(self):
        """
        Returns whether the service is running or not.

        Returns:
            Boolean: `True` if service is running, `False` otherwise.
        """
        return self.service_active

    def get_interface_type(self):
        """
        Returns the interface type.

        Returns:
            :class:`.BLEInterface`: The interface type.
        """
        return self._interface_type

    def add_data_received_callback(self, callback):
        """
        Adds a new callback to the list of callbacks that will be notified when data is received
        from the connected device.

        Args:
            callback (Function): the new callback function.
        """
        self._on_data_received.append(callback)

    def remove_data_received_callback(self, callback):
        """
        Removes the given callback from the `data_received` callbacks list.

        Args:
            callback (Function): the callback function to be removed.
        """
        if callback in self._on_data_received:
            self._on_data_received.remove(callback)

    def add_connect_callback(self, callback):
        """
        Adds a new callback to the list of callbacks that will be notified when a device connects.

        Args:
            callback (Function): the new callback function.
        """
        self._on_connect.append(callback)

    def remove_connect_callback(self, callback):
        """
        Removes the given callback from the list of `connect` callbacks.

        Args:
            callback (Function): the callback function to be removed.
        """
        if callback in self._on_connect:
            self._on_connect.remove(callback)

    def add_disconnect_callback(self, callback):
        """
        Adds a new callback to the list of callbacks that will be notified when a device
        disconnects.

        Args:
            callback (Function): the new callback function.
        """
        self._on_disconnect.append(callback)

    def remove_disconnect_callback(self, callback):
        """
        Removes the given callback from the list of `disconnect` callbacks.

        Args:
            callback (Function): the callback function to be removed.
        """
        if callback in self._on_disconnect:
            self._on_disconnect.remove(callback)

    @abstractmethod
    def start(self):
        """
        Starts the BLE service.

        Raises:
            ConnectCoreBLEException: if an error occurs while starting the service.
        """
        raise NotImplementedError()

    @abstractmethod
    def stop(self):
        """
        Stops the BLE service.

        Raises:
            ConnectCoreBLEException: if an error occurs while stopping the service.
        """
        raise NotImplementedError()

    @abstractmethod
    def send_data(self, data):
        """
        Sends the given data to the connected device through the available BLE interface.

        Args:
            data (Bytearray): data to send.

        Raises:
            ConnectCoreBLEException: if an error occurs while sending the data.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_advertising_name(self):
        """
        Changes the currently advertised service name.

        Returns:
            String: the current advertising name.

        Raises:
            ConnectCoreBLEException: if an error occurs while reading the advertising name.
        """
        raise NotImplementedError()

    @abstractmethod
    def set_advertising_name(self, device_name):
        """
        Changes the currently advertised service name.

        Args:
            device_name (String): the new name of the service.

        Raises:
            ConnectCoreBLEException: if an error occurs while changing the advertising name.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_device_connected(self):
        """
        Checks whether there is any device connected to the BLE interface or not.

        Returns:
            Boolean: `True` if there is any device connected, `False` otherwise.
        """
        raise NotImplementedError()

    @abstractmethod
    def set_password(self, new_password):
        """
        Sets the new authentication password for the service.

        Args:
            new_password (String): the new authentication password.

        Raises:
            ConnectCoreBLEException: if an error occurs while setting the new password.
        """
        raise NotImplementedError()


class ConnectCoreBLEServiceNative(ConnectCoreBLEService):
    """
    This class represents a native `ConnectCoreBLEService` by creating a local GATT server with an
    UART peripheral exposing read and write interfaces.

    The service uses SRP authentication protocol to encrypt and decrypt data with a default
    password of '1234'.
    """

    # Constants.
    _DEFAULT_AUTHENTICATION_KEY = "1234"
    _DEFAULT_FIRMWARE_VERSION = [0, 0, 0x10, 0x0D]

    def __init__(self, adapter_address):
        """
        Class constructor. Instantiates a new :class:`.ConnectCoreBLEServiceNative` object with the
        given parameters.

        Args:
            adapter_address (String): the address of the native Bluetooth adapter to use.
        """
        ConnectCoreBLEService.__init__(self, BLEInterface.ADAPTER)

        # Instantiate peripheral.
        self._peripheral = ConnectCoreBLEPeripheral(adapter_address)
        # Set callback functions.
        self._peripheral.add_data_received_callback(self._data_received)
        self._peripheral.add_connection_changed_callback(self._connection_changed)
        self._security_manager = SRPSecurityManager()
        # Initialize SRP authentication.
        self._security_manager.generate_salted_verification_key(self._DEFAULT_AUTHENTICATION_KEY)

    def start(self):
        """
        Override.

        See Also:
            :meth:`.ConnectCoreBLEService.start`
        """
        # Sanity checks.
        if self.service_active:
            return

        try:
            self.service_active = True
            self._peripheral.start()
        except Exception as exc:
            self.service_active = False
            self._log.error(_ERROR_START_BLUETOOTH_SERVICE, str(exc))
            raise ConnectCoreBLEException(_ERROR_START_BLUETOOTH_SERVICE % str(exc)) from exc

    def stop(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.stop`
        """
        # Sanity checks.
        if not self.service_active:
            return

        self._peripheral.stop()
        self.service_active = False

    def send_data(self, data):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.send_data`
        """
        # Initialize variables.
        encrypt = True
        # If the packet is a Bluetooth Unlock API Frame answer, prepare the SRP response.
        # This packet must not be encrypt as it is part of the SRP handshake.
        if len(data) > 3 and data[3] == _XBEE_API_FRAME_SRP_LOGIN_ANSWER:
            packet = UnknownXBeePacket.create_packet(data)
            data_to_send = packet.output()
            encrypt = False
        elif len(data) > 3 and data[3] == ApiFrameType.AT_COMMAND_RESPONSE.code:
            # If the packet is an AT Command Answer, just send the given data.
            data_to_send = data
        else:
            # For any other data, build an XBee User Relay Output frame.
            packet = UserDataRelayOutputPacket(XBeeLocalInterface.BLUETOOTH, data)
            data_to_send = packet.output()
        # Encrypt the data if required.
        if encrypt:
            try:
                data_to_send = self._security_manager.encrypt_data(data_to_send)
            except ConnectCoreBLEException as exc:
                self._log.warning(_WARNING_DATA_ENCRYPT, str(exc))
                return
        # Try to send the data.
        try:
            self._peripheral.send_rx_data(data_to_send)
        except DBusException as exc:
            self._log.error(_ERROR_SEND_DATA, str(exc))
            raise ConnectCoreBLEException(_ERROR_SEND_DATA % str(exc)) from exc

    def get_advertising_name(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.get_advertising_name`
        """
        return self._peripheral.get_advertising_name()

    def set_advertising_name(self, device_name):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.set_advertising_name`
        """
        try:
            self._peripheral.configure_advertising_name(device_name)
        except Exception as exc:
            self._log.error(_ERROR_CHANGE_ADVERTISE_NAME, str(exc))
            raise ConnectCoreBLEException(_ERROR_CHANGE_ADVERTISE_NAME % str(exc)) from exc

    def is_device_connected(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.is_device_connected`
        """
        return self._peripheral.is_device_connected()

    def set_password(self, new_password):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.set_password`
        """
        try:
            self._security_manager.generate_salted_verification_key(new_password)
        except ValueError as exc:
            self._log.error(_ERROR_CHANGE_ENCRYPTION_PASSWORD, str(exc))
            raise ConnectCoreBLEException(_ERROR_CHANGE_ENCRYPTION_PASSWORD % str(exc)) from exc

    def _data_received(self, data):
        """
        Callback function to be notified when new data is received from the device connected
        to the peripheral.

        Args:
            data (Bytearray): the data received from the device connected to the peripheral.
        """
        # If the packet is a Bluetooth Unlock API Frame, process the SRP request and return.
        # This process is executed before SRP encryption is set, so do not decode data here and
        # execute only if user is not authenticated yet.
        if (len(data) > 3 and data[3] == _XBEE_API_FRAME_SRP_LOGIN_REQUEST):
            try:
                XBeeAPIPacket._check_api_packet(data)
                self._security_manager.deauthenticate()
                srp_response = self._security_manager.process_srp_request(data)
                self.send_data(srp_response)
                return
            except InvalidPacketException:
                self._log.warning(_WARNING_SPURIOUS_LOGIN_PACKET_RECEIVED)
        # Decrypt the data if possible.
        try:
            decrypted_data = self._security_manager.decrypt_data(data)
        except ConnectCoreBLEException as exc:
            self._log.warning(_WARNING_DATA_DECRYPT, str(exc))
            return
        # Check if it is an AT Command request to obtain device information.
        if len(decrypted_data) > 3 and decrypted_data[3] == ApiFrameType.AT_COMMAND.code:
            # Get the AT Command request answer.
            answer_packet = self._process_at_command_request(decrypted_data)
            # If the answer is valid, send it to the device, otherwise the given packet was invalid.
            if answer_packet is not None:
                self.send_data(answer_packet.output())
            return
        # Check if it is a User Data Relay Request frame.
        if (len(decrypted_data) > 3
                and decrypted_data[3] == ApiFrameType.USER_DATA_RELAY_REQUEST.code):
            # Try to create an XBee user data relay frame with the given data.
            try:
                packet = UserDataRelayPacket.create_packet(decrypted_data, OperatingMode.API_MODE)
                # Notify subscribed service callbacks about new data received.
                for callback in self._on_data_received:
                    callback(packet.data)
            except InvalidPacketException as exc:
                self._log.warning(_WARNING_INVALID_PACKET_RECEIVED, str(exc))
                return

    def _connection_changed(self, status):
        """
        Callback function to be notified when a connectivity change occurs in the BLE peripheral.

        Args:
            status (Boolean): `True` if any device connects, `False` if any device disconnects.
        """
        # Notify corresponding registered service callbacks.
        if status:
            for callback in self._on_connect:
                callback()
        else:
            for callback in self._on_disconnect:
                callback()

    def _process_at_command_request(self, data):
        """
        Processes the given data as an AT Command Request and returns the requested device
        information in an AT Command Response packet.

        Args:
            data (Bytearray): Data to process.

        Returns:
            :class:`.ATCommResponsePacket`: the AT Command Response packet, `None` if the
                                            given data is invalid.
        """
        # Initialize variables.
        payload = bytearray()
        # Try to create an XBee user data relay frame with the given data.
        try:
            packet = ATCommPacket.create_packet(data, OperatingMode.API_MODE)
            if packet.command == ATStringCommand.SH.command:
                # Use 0x0000 plus the first 2 bytes of the adapter address.
                payload = bytearray.fromhex("0000".join(
                    self._peripheral.dongle.address.split(":")[0:2]))
            elif packet.command == ATStringCommand.SL.command:
                # Use last 4 bytes of the adapter address.
                payload = bytearray.fromhex("".join(
                    self._peripheral.dongle.address.split(":")[2:]))
            elif packet.command == ATStringCommand.NI.command:
                # Use advertising name as device identifier.
                payload = bytearray(self._peripheral.local_name.encode(encoding="utf-8"))
            elif packet.command == ATStringCommand.HV.command:
                # Report ourselves as an XBee 3 SMT
                payload = bytearray([HardwareVersion.XBEE3_SMT.code, 0])
            elif packet.command == ATStringCommand.VR.command:
                payload = bytearray(self._DEFAULT_FIRMWARE_VERSION)
            elif packet.command == ATStringCommand.MY.command:
                # Use the coordinator address '0000'
                payload = bytearray([0, 0])
            # If the payload is filled then the command was valid.
            if len(payload) > 0:
                response_status = ATCommandStatus.OK
            else:
                response_status = ATCommandStatus.INVALID_COMMAND
            # Build and return the response packet.
            return ATCommResponsePacket(packet.frame_id,
                                        packet.command,
                                        response_status=response_status,
                                        comm_value=payload,
                                        op_mode=OperatingMode.API_MODE)
        except InvalidPacketException as exc:
            self._log.warning(_WARNING_INVALID_PACKET_RECEIVED, str(exc))
            return None


class BLEServiceXBee(ConnectCoreBLEService):
    """
    This class implements the abstract methods of `ConnectCoreBLEService` by establishing
    connection with an XBee device and communicating with its GATT server.

    The service uses SRP authentication protocol to encrypt and decrypt data managed by the XBee
    device.
    """

    def __init__(self, open_xbee):
        ConnectCoreBLEService.__init__(self, BLEInterface.XBEE3)

        # Initialize variables.
        self._stop_service_event = threading.Event()
        self._xbee = open_xbee
        self._connected = False
        # Subscribe callbacks callback.
        self._xbee.add_bluetooth_data_received_callback(self._data_received)
        self._xbee.add_modem_status_received_callback(self._modem_status_changed)

    def start(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.start`
        """
        # Sanity checks.
        if self.service_active:
            return

        try:
            # Enable bluetooth service in the XBee device.
            self._xbee.enable_bluetooth()
        except XBeeException as exc:
            self._log.error(_ERROR_START_BLUETOOTH_SERVICE, str(exc))
            raise ConnectCoreBLEException(_ERROR_START_BLUETOOTH_SERVICE % str(exc)) from exc
        # Wait for new connections.
        wait_thread = threading.Thread(target=self._stop_service_event.wait)
        wait_thread.start()
        # Set active flag.
        self.service_active = True

    def stop(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.stop`
        """
        # Sanity checks.
        if not self.service_active:
            return

        try:
            # Disable bluetooth service in the XBee device.
            self._xbee.disable_bluetooth()
        except XBeeException as exc:
            self._log.error(_ERROR_STOP_BLUETOOTH_SERVICE, str(exc))
            raise ConnectCoreBLEException(_ERROR_STOP_BLUETOOTH_SERVICE % str(exc)) from exc
        # Release wait thread.
        self._stop_service_event.set()
        # Set active flag.
        self.service_active = False

    def send_data(self, data):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.send_data`
        """
        try:
            self._xbee.send_bluetooth_data(data)
        except XBeeException as exc:
            self._log.error(_ERROR_SEND_DATA, str(exc))
            raise ConnectCoreBLEException(_ERROR_SEND_DATA % str(exc)) from exc

    def get_advertising_name(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.get_advertising_name`
        """
        try:
            return self._xbee.get_parameter(ATStringCommand.BI.command).decode(encoding="utf-8")
        except XBeeException as exc:
            self._log.error(_ERROR_READ_ADVERTISE_NAME, str(exc))
            raise ConnectCoreBLEException(_ERROR_READ_ADVERTISE_NAME % str(exc)) from exc

    def set_advertising_name(self, device_name):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.set_advertising_name`
        """
        try:
            self._xbee.set_parameter(ATStringCommand.BI.command,
                                     bytearray(device_name, "utf-8"),
                                     apply=True)
        except XBeeException as exc:
            self._log.error(_ERROR_CHANGE_ADVERTISE_NAME, str(exc))
            raise ConnectCoreBLEException(_ERROR_CHANGE_ADVERTISE_NAME % str(exc)) from exc

    def is_device_connected(self):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.is_device_connected`
        """
        return self._connected

    def set_password(self, new_password):
        """
        Override.

        .. seealso::
           | :meth:`.ConnectCoreBLEService.set_password`
        """
        try:
            self._xbee.update_bluetooth_password(new_password)
        except (ValueError, XBeeException) as exc:
            self._log.error(_ERROR_CHANGE_ENCRYPTION_PASSWORD, str(exc))
            raise ConnectCoreBLEException(_ERROR_CHANGE_ENCRYPTION_PASSWORD % str(exc)) from exc

    def _data_received(self, data):
        """
        Callback function to be notified when new data is received from the device
        connected to the XBee.

        Args:
            data (Bytearray): the data received from the device connected to the XBee.
        """
        # Notify corresponding registered service callbacks.
        for callback in self._on_data_received:
            callback(data)

    def _modem_status_changed(self, modem_status):
        """
        Callback function to be notified when the XBee modem status changes.

        Args:
            modem_status (:class:`.ModemStatus`): the new modem status.
        """
        if modem_status == ModemStatus.BLUETOOTH_CONNECTED:
            # A device just connected to the XBee.
            self._connected = True
            # Notify corresponding registered service callbacks.
            for callback in self._on_connect:
                callback()
        elif modem_status == ModemStatus.BLUETOOTH_DISCONNECTED:
            # A device just disconnected from the XBee.
            self._connected = False
            # Notify corresponding registered service callbacks.
            for callback in self._on_disconnect:
                callback()



# Copyright 2022, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging

from bluezero.peripheral import Peripheral
from dbus.exceptions import DBusException
from digi.ccble import utils

# Constants.
CHAR_RX_ID = 1
CHAR_RX_UUID = "F9279EE9-2CD0-410C-81CC-ADF11E4E5AEA"
CHAR_TX_ID = 2
CHAR_TX_UUID = "7DDDCA00-3E05-4651-9254-44074792C590"
PERIPHERAL_NAME = "CONNECTCORE_{}"
SERVICE_ID = 1
SERVICE_UUID = "53DA53B9-0447-425A-B9EA-9837505EB59A"


class ConnectCoreBLEPeripheral(Peripheral):
    """
    Class used to represent a Bluetooth Low Energy UART peripheral.

    The peripheral offers a UART service with two characteristics:
      - TX Characteristic: Used to allow connected devices to send data.
      - RX Characteristic: Used to allow connected devices to read/receive data.
    """

    # Variables.
    _log = logging.getLogger(__name__)

    def __init__(self, adapter_address):
        """
        Class constructor. Instantiates a new :class:`.ConnectCoreBLEPeripheral` object
        with the given adapter address.

        Args:
            adapter_address (String): local adapter address.
        """
        # Call Peripheral constructor.
        Peripheral.__init__(self,
                            adapter_address,
                            local_name=PERIPHERAL_NAME.format(
                                "".join(adapter_address.split(":")[-2:])))

        # Initialize variables.
        self._rx_characteristic = None
        self._running = False
        self._connected = False
        self._data_received_callbacks = []
        self._connection_changed_callbacks = []

        # Create UART service.
        self.add_service(srv_id=SERVICE_ID, uuid=SERVICE_UUID, primary=True)
        # Create RX characteristic (available data for the device)
        self.add_characteristic(srv_id=SERVICE_ID,
                                chr_id=CHAR_RX_ID,
                                uuid=CHAR_RX_UUID,
                                value=[],
                                notifying=False,
                                flags=['notify'],
                                read_callback=None,
                                write_callback=None,
                                notify_callback=self._notify_changed)
        # Create TX characteristic (incoming data from the device)
        self.add_characteristic(srv_id=SERVICE_ID,
                                chr_id=CHAR_TX_ID,
                                uuid=CHAR_TX_UUID,
                                value=[],
                                notifying=False,
                                flags=['write', 'write-without-response'],
                                read_callback=None,
                                write_callback=self._data_written,
                                notify_callback=None)
        # Register connection callbacks.
        self.on_connect = self._device_connected
        self.on_disconnect = self._device_disconnected

    def _data_written(self, value, _options):
        """
        Callback function to be notified when new data is written to the
        TX characteristic by the device.

        Args:
            value (Byterray): data written to the TX characteristic.
            _options (Dict): options used to write the data.
        """
        self._log.debug("Data written: %s", utils.hex_to_string(value))
        # Notify callbacks if any.
        for callback in self._data_received_callbacks:
            callback(value)

    def _notify_changed(self, notifying, characteristic):
        """
        Callback function to be notified when the notify state of the RX
        characteristic changes.

        Args:
            notifying (Boolean): `True` if the RX characteristic is notifying,
                                 `False` otherwise.
            characteristic (:class:`.GATTCharacteristic`): RX characteristic.
        """
        self._log.info("RX notify state changed: %s", notifying)
        # When the notify state changes, update the RX characteristic.
        if notifying:
            self._rx_characteristic = characteristic
        else:
            self._rx_characteristic = None

    def _device_connected(self, device=None, _address=None):
        """
        Callback function to be notified when a device connects.

        Args:
            device (:class:`.Device`): device instance of the connected target.
            _address (String): local adapter address followed by the remote address.
        """
        self._log.info("Device connected" if device is None else "Device connected: %s", device)
        self._connected = True
        # Notify callbacks if any.
        for callback in self._connection_changed_callbacks:
            callback(True)

    def _device_disconnected(self, device=None, _address=None):
        """
        Callback function to be notified when a device disconnects.

        Args:
            device (:class:`.Device`): device instance of the disconnected target.
            _address (String): local adapter address followed by the remote address.
        """
        self._log.info("Device disconnected" if device is None else
                       "Device disconnected: %s", device)
        self._connected = False
        # Notify callbacks if any.
        for callback in self._connection_changed_callbacks:
            callback(False)

    def start(self):
        """
        Starts the peripheral.
        """
        self._log.info("Starting peripheral '%s'...", self.local_name)
        self._running = True
        try:
            self.publish()
        except Exception as exc:
            self._log.error("Error starting peripheral: %s", str(exc))
            self._running = False
            raise

    def stop(self):
        """
        Stops the peripheral.
        """
        self.mainloop.quit()
        try:
            self.ad_manager.unregister_advertisement(self.advert)
        except DBusException:
            # Ignore this error, service has stopped.
            pass
        self._running = False
        self._log.info("Peripheral stopped")

    def send_rx_data(self, data):
        """
        Sends new data to the RX Characteristic.

        Args:
            data (Byterray): data to be sent to the RX Characteristic.

        Raises:
            DBusException: if there is any error writing in the RX characteristic.
        """
        if self._rx_characteristic is not None:
            self._rx_characteristic.set_value(data)
            self._log.debug("Sent data: %s", utils.hex_to_string(data))

    def get_advertising_name(self):
        """
        Returns the advertising device name.

        Returns:
            String: the advertising name.
        """
        return self.local_name

    def configure_advertising_name(self, device_name):
        """
        Modifies the advertising device name.

        Args:
            device_name (String): new advertising name.
        """
        self._log.info("Configuring advertising name to '%s'", device_name)
        self.local_name = device_name
        if not self._connected and self._running:
            self.stop()
            self.start()

    def is_device_connected(self):
        """
        Returns whether there is any device connected or not.

        Returns:
            Boolean: `True` if there is any device connected, `False` otherwise.
        """
        return self._connected

    def add_connection_changed_callback(self, callback):
        """
        Adds a new callback to be notified when the connection status changes.

        Args:
            callback (Function): the callback function to add. Receives a `Boolean` object.
        """
        if callback not in self._connection_changed_callbacks:
            self._connection_changed_callbacks.append(callback)

    def del_connection_changed_callback(self, callback):
        """
        Removes the given callback from the list.

        Args:
            callback (Function): the callback function to remove.
        """
        if callback in self._connection_changed_callbacks:
            self._connection_changed_callbacks.remove(callback)

    def add_data_received_callback(self, callback):
        """
        Adds a new callback to be notified when data is written to the TX characteristic.

        Args:
            callback (Function): the callback function to add. Receives a `Byterray` object.
        """
        if callback not in self._data_received_callbacks:
            self._data_received_callbacks.append(callback)

    def del_data_received_callback(self, callback):
        """
        Removes the given callback from the list.

        Args:
            callback (Function): the callback function to remove.
        """
        if callback in self._data_received_callbacks:
            self._data_received_callbacks.remove(callback)



# Copyright 2022, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os

from enum import Enum, unique

import srp

from Crypto.Cipher import AES
from Crypto.Util import Counter
from digi.ccble import utils
from digi.ccble.exceptions import NotAuthenticatedException
from digi.xbee.packets.base import UnknownXBeePacket

@unique
class SrpPhase(Enum):
    """
    Enumeration class listing all the SRP phases.
    """
    PHASE_1 = (0x01, "Phase 1: Client presents A value")
    PHASE_2 = (0x02, "Phase 2: Server presents B and salt")
    PHASE_3 = (0x03, "Phase 3: Client presents M1 session key validation value")
    PHASE_4 = (0x04, "Phase 4: Server presents M2 session key validation value and two 12-byte nonces")
    UNKNOWN = (0xFF, "Unknown")

    def __init__(self, code, description):
        """
        Class constructor. Instantiates a new `SrpPhase` entry with the provided parameters.

        Args:
            code (Integer): SRP phase code.
            description (String): SRP phase description.
        """
        self.__code = code
        self.__description = description

    @property
    def code(self):
        """
        Returns the SRP phase code.

        Returns:
            Integer: SRP phase code.
        """
        return self.__code

    @property
    def description(self):
        """
        Returns the SRP phase description.

        Returns:
            String: SRP phase description.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the SRP phase corresponding to the given code.

        Args:
            code (Integer): SRP phase code.

        Returns:
            SrpPhase: SRP phase corresponding to the given code.
        """
        for srp_phase in cls:
            if code == srp_phase.code:
                return srp_phase
        return SrpPhase.UNKNOWN


@unique
class SrpError(Enum):
    """
    Enumeration class listing all the SRP errors.
    """
    B_OFFERING_ERROR = (0x80, "Unable to offer B (cryptographic error with content, usually due to A mod N == 0")
    INCORRECT_LENGTH = (0x81, "Incorrect payload length")
    BAD_PROOF_OF_KEY = (0x82, "Bad proof of key")
    ALLOCATION_ERROR = (0x83, "Resource allocation error")
    WRONG_STEP_ERROR = (0x84, "Request contained a step not in the correct sequence")

    def __init__(self, code, description):
        """
        Class constructor. Instantiates a new `SrpError` entry with the provided parameters.

        Args:
            code (Integer): SRP error code.
            description (String): SRP error description.
        """
        self.__code = code
        self.__description = description

    @property
    def code(self):
        """
        Returns the SRP error code.

        Returns:
            Integer: SRP error code.
        """
        return self.__code

    @property
    def description(self):
        """
        Returns the SRP error description.

        Returns:
            String: SRP error description.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the SRP error corresponding to the given code.

        Args:
            code (Integer): SRP error code.

        Returns:
            SrpError: SRP error corresponding to the given code, `None` if not found.
        """
        for srp_error in cls:
            if code == srp_error.code:
                return srp_error
        return None


SrpPhase.__doc__ += utils.doc_enum(SrpPhase)
SrpError.__doc__ += utils.doc_enum(SrpError)


class SRPSecurityManager:
    """
    Helper class used to handle SRP authentication protocol with a remote device in order to encrypt
    and decrypt data.
    """

    # Constants.
    _DEFAULT_USERNAME = 'apiservice'

    def __init__(self):
        """
        Class constructor. Instantiates a new `SRPSecurityManager` object.
        """
        # Initialize variables.
        self._salt = None
        self._verification_key = None
        self._verifier = None
        self._encryptor = None
        self._decryptor = None
        self._authenticated = False

    def is_authenticated(self):
        """
        Returns whether the session is authenticated or not.

        Returns:
            Boolean: `True` if the session is authenticated, `False` otherwise.
        """
        return self._authenticated

    def deauthenticate(self):
        """
        Deauthenticate the security manager.
        """
        self._authenticated = False

    @staticmethod
    def new_cipher(session_key, nonce):
        """
        Creates a new AES cipher with the given session key and nonce.

        Args:
            session_key (Bytearray): Session key.
            nonce (Bytearray): Nonce.

        Returns:
            :class:`.AES`: AES cipher.
        """
        counter = Counter.new(nbits=32, prefix=nonce, initial_value=1)
        cipher = AES.new(key=session_key, mode=AES.MODE_CTR, counter=counter)
        return cipher

    def encrypt_data(self, data):
        """
        Encrypts the given data with the stored encrypt cipher.

        Args:
            data (Bytearray): Data to be encrypt.

        Returns:
            Bytearray: Encrypted data.

        Raises:
            NotAuthenticatedException: If the user is not authenticated.
        """
        # Sanity checks.
        if not self._authenticated:
            raise NotAuthenticatedException()

        return bytearray(self._encryptor.encrypt(bytes(data)))

    def decrypt_data(self, data):
        """
        Decrypts the given data with the stored decrypt cipher.

        Args:
            data (Bytearray): Data to be decrypt.

        Returns:
            Bytearray: Decrypted data.

        Raises:
            NotAuthenticatedException: If the user is not authenticated.
        """
        # Sanity checks.
        if not self._authenticated:
            raise NotAuthenticatedException()

        return bytearray(self._decryptor.decrypt(bytes(data)))

    def generate_salted_verification_key(self, password):
        """
        Generates and stores a salted verification key from the given password.

        Args:
            password (String): Password to use to generate the salted verification key.
        """
        self._salt, self._verification_key = srp.create_salted_verification_key(self._DEFAULT_USERNAME,
                                                                                password,
                                                                                hash_alg=srp.SHA256,
                                                                                ng_type=srp.NG_1024,
                                                                                salt_len=4)

    def process_srp_request(self, frame_data):
        """
        Processes the SRP request contained in the given data and returns an XBee packet with the
        answer.

        Args:
            frame_data (bytes): Data containing the SRP request.

        Returns:
            :class:`.XBeeAPIPacket`: XBee packet containing the SRP answer, `None` if the request is
                                     not an SRP request.
        """
        srp_phase = SrpPhase.get(frame_data[4])
        if srp_phase == SrpPhase.PHASE_1:
            # Phase 1: Client presents 'A' value, we have to present 'B' value and salt.
            # --------------------------------------------------------------------------
            # Initialize payload.
            payload = SrpPhase.PHASE_2.code.to_bytes(1, byteorder='big')
            # Extract client 'ephemeral a' from the given data.
            client_ephemeral_a = bytes.fromhex(utils.hex_to_string(frame_data[5:133]))
            # Generate SRP verifier with the stored salted key and the ephemeral_a.
            self._verifier = srp.Verifier(self._DEFAULT_USERNAME,
                                          self._salt,
                                          self._verification_key,
                                          client_ephemeral_a,
                                          hash_alg=srp.SHA256,
                                          ng_type=srp.NG_1024)
            # Calculate server 'ephemeral b'.
            _, server_ephemeral_b = self._verifier.get_challenge()
            # Check for error.
            if server_ephemeral_b is None:
                # Add error code to payload.
                payload = SrpError.B_OFFERING_ERROR.code
            else:
                # Add server 'ephemeral b' and 'salt' to payload.
                payload += self._salt
                payload += server_ephemeral_b
            # Generate and return an XBee packet with the payload.
            return UnknownXBeePacket(0xAC, bytearray(payload)).output()
        if srp_phase == SrpPhase.PHASE_3:
            # Phase 3: Client presents 'M1 session key validation value,' we have to present
            # 'M2 session key validation value' and two 12-byte 'nonces'.
            # ------------------------------------------------------------------------------
            # Initialize payload.
            payload = SrpPhase.PHASE_4.code.to_bytes(1, byteorder='big')
            # Extract 'client proof m1' from the given data.
            client_proof_m1 = bytes.fromhex(utils.hex_to_string(frame_data[5:37]))
            # Calculate 'server proof m2'.
            server_proof_m2 = self._verifier.verify_session(client_proof_m1)
            # Initialize key.
            key = None
            # Check for error.
            if server_proof_m2 is None or not self._verifier.authenticated():
                # Add error code to payload.
                payload = SrpError.BAD_PROOF_OF_KEY.code.to_bytes(1, byteorder='big')
            else:
                # Generate two 12-byte random 'nonces'.
                tx_nonce = os.urandom(12)
                rx_nonce = os.urandom(12)
                # Fill payload.
                payload += server_proof_m2
                payload += tx_nonce
                payload += rx_nonce
                # Extract session key.
                key = self._verifier.get_session_key()
                # Create encryptor and decryptor using the session key and generated 'nonces'.
                self._encryptor = self.new_cipher(key, rx_nonce)
                self._decryptor = self.new_cipher(key, tx_nonce)
            # Check if key generation succeed.
            if key is not None:
                self._authenticated = True
            # Generate and return an XBee packet with the payload.
            return UnknownXBeePacket(0xAC, bytearray(payload)).output()

        # Invalid SRP phase or not SRP request packet.
        return None



# Copyright 2022, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import sys

from codecs import open
from setuptools import setup, find_namespace_packages

# 'setup.py build' shortcut.
if sys.argv[-1] == 'build':
    os.system('python setup.py sdist bdist_wheel')
    sys.exit()

# 'setup.py publish' shortcut, uses the .pypirc in your home directory.
if sys.argv[1] == 'publish':
    repo = " -r %s" % sys.argv[2] if len(sys.argv) > 2 else ""
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload%s dist/*' % repo)
    sys.exit()

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, 'digi', 'ccble', '__init__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

requirements = [i.strip() for i in open("requirements.txt").readlines()]

setup(
    name=about['__title__'],
    namespace_packages=['digi'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    packages=find_namespace_packages(include=['digi.*']),
    package_data={'': ['LICENSE.txt']},
    include_package_data=True,
    keywords=['bluetooth', 'connectcore', 'ble', 'digi', 'xbee'],
    license=about['__license__'],
    python_requires=">=3, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, !=3.7.*",
    install_requires=requirements,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Software Development :: Libraries',
        'Topic :: Home Automation',
        'Topic :: Games/Entertainment',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],
    project_urls={
        'Documentation': 'https://digi-connectcore-ble.readthedocs.io/en/latest/',
        'Source': 'https://github.com/digi-embedded/connectcore-ble-python',
        'Tracker': 'https://github.com/digi-embedded/connectcore-ble-python/issues',
    },
)
